#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

import unittest
import requests

from gallery_dl import text
from gallery_dl.util import OAuthSession

TESTSERVER = "http://oauthbin.com"
CONSUMER_KEY = "key"
CONSUMER_SECRET = "secret"
REQUEST_TOKEN = "requestkey"
REQUEST_TOKEN_SECRET = "requestsecret"
ACCESS_TOKEN = "accesskey"
ACCESS_TOKEN_SECRET = "accesssecret"


class TestOAuthSession(unittest.TestCase):

    def test_concat(self):
        concat = OAuthSession.concat

        self.assertEqual(concat(), "")
        self.assertEqual(concat("str"), "str")
        self.assertEqual(concat("str1", "str2"), "str1&str2")

        self.assertEqual(concat("&", "?/"), "%26&%3F%2F")
        self.assertEqual(
            concat("GET", "http://example.org/", "foo=bar&baz=a"),
            "GET&http%3A%2F%2Fexample.org%2F&foo%3Dbar%26baz%3Da"
        )

    def test_nonce(self, N=16):
        nonce_values = set(OAuthSession.nonce(N) for _ in range(N))

        # uniqueness
        self.assertEqual(len(nonce_values), N)

        # length
        for nonce in nonce_values:
            self.assertEqual(len(nonce), N)

    def test_quote(self):
        quote = OAuthSession.quote

        reserved = ",;:!\"§$%&/(){}[]=?`´+*'äöü"
        unreserved = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                      "abcdefghijklmnopqrstuvwxyz"
                      "0123456789-._~")

        for char in unreserved:
            self.assertEqual(quote(char), char)

        for char in reserved:
            quoted = quote(char)
            quoted_hex = quoted.replace("%", "")
            self.assertTrue(quoted.startswith("%"))
            self.assertTrue(len(quoted) >= 3)
            self.assertEqual(quoted_hex.upper(), quoted_hex)

    def test_urlencode(self):
        urlencode = OAuthSession.urlencode

        self.assertEqual(urlencode({}), "")
        self.assertEqual(urlencode({"foo": "bar"}), "foo=bar")

        self.assertEqual(
            urlencode({"foo": "bar", "baz": "a", "a": "baz"}),
            "a=baz&baz=a&foo=bar"
        )
        self.assertEqual(
            urlencode({
                "oauth_consumer_key": "0685bd9184jfhq22",
                "oauth_token": "ad180jjd733klru7",
                "oauth_signature_method": "HMAC-SHA1",
                "oauth_timestamp": 137131200,
                "oauth_nonce": "4572616e48616d6d65724c61686176",
                "oauth_version": "1.0"
            }),
            "oauth_consumer_key=0685bd9184jfhq22&"
            "oauth_nonce=4572616e48616d6d65724c61686176&"
            "oauth_signature_method=HMAC-SHA1&"
            "oauth_timestamp=137131200&"
            "oauth_token=ad180jjd733klru7&"
            "oauth_version=1.0"
        )

    def test_request_token(self):
        response = self._oauth_request(
            "/v1/request-token", {})
        expected = "oauth_token=requestkey&oauth_token_secret=requestsecret"
        self.assertEqual(response, expected, msg=response)

        data = text.parse_query(response)
        self.assertTrue(data["oauth_token"], REQUEST_TOKEN)
        self.assertTrue(data["oauth_token_secret"], REQUEST_TOKEN_SECRET)

    def test_access_token(self):
        response = self._oauth_request(
            "/v1/access-token", {}, REQUEST_TOKEN, REQUEST_TOKEN_SECRET)
        expected = "oauth_token=accesskey&oauth_token_secret=accesssecret"
        self.assertEqual(response, expected, msg=response)

        data = text.parse_query(response)
        self.assertTrue(data["oauth_token"], ACCESS_TOKEN)
        self.assertTrue(data["oauth_token_secret"], ACCESS_TOKEN_SECRET)

    def test_authenticated_call(self):
        params = {"method": "foo", "bar": "baz", "a": "äöüß/?&#"}
        response = self._oauth_request(
            "/v1/echo", params, ACCESS_TOKEN, ACCESS_TOKEN_SECRET)
        expected = OAuthSession.urlencode(params)

        self.assertEqual(response, expected, msg=response)
        self.assertEqual(text.parse_query(response), params)

    def _oauth_request(self, endpoint, params=None,
                       oauth_token=None, oauth_token_secret=None):
        session = OAuthSession(
            requests.session(),
            CONSUMER_KEY, CONSUMER_SECRET,
            oauth_token, oauth_token_secret,
        )
        url = TESTSERVER + endpoint
        return session.get(url, params.copy()).text


if __name__ == "__main__":
    unittest.main(warnings="ignore")



# -*- coding: utf-8 -*-

# Copyright 2016-2018 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extract images from https://www.pinterest.com"""

from .common import Extractor, Message
from .. import text, util, exception


class PinterestExtractor(Extractor):
    """Base class for pinterest extractors"""
    category = "pinterest"
    filename_fmt = "{category}_{pin_id}.{extension}"
    archive_fmt = "{pin_id}"

    def __init__(self):
        Extractor.__init__(self)
        self.api = PinterestAPI(self)

    def data_from_pin(self, pin):
        """Get image url and metadata from a pin-object"""
        img = pin["image"]["original"]
        url = img["url"]
        data = {
            "pin_id": util.safe_int(pin["id"]),
            "note": pin["note"],
            "width": util.safe_int(img["width"]),
            "height": util.safe_int(img["height"]),
        }
        return url, text.nameext_from_url(url, data)


class PinterestPinExtractor(PinterestExtractor):
    """Extractor for images from a single pin from pinterest.com"""
    subcategory = "pin"
    pattern = [r"(?:https?://)?(?:[^./]+\.)?pinterest\.[^/]+/pin/([^/?#&]+)"]
    test = [
        ("https://www.pinterest.com/pin/858146903966145189/", {
            "url": "afb3c26719e3a530bb0e871c480882a801a4e8a5",
            "keyword": "f651cb271247f306d1d30385d49c7b82da44c2b1",
            "content": "d3e24bc9f7af585e8c23b9136956bd45a4d9b947",
        }),
        ("https://www.pinterest.com/pin/858146903966145188/", {
            "exception": exception.StopExtraction,
        }),
        ("https://www.pinterest.com/pin/85814690396614518/", {
            "exception": exception.NotFoundError,
        }),
    ]

    def __init__(self, match):
        PinterestExtractor.__init__(self)
        self.pin_id = match.group(1)

    def items(self):
        pin = self.api.pin(self.pin_id)
        url, data = self.data_from_pin(pin)
        yield Message.Version, 1
        yield Message.Directory, data
        yield Message.Url, url, data


class PinterestBoardExtractor(PinterestExtractor):
    """Extractor for images from a board from pinterest.com"""
    subcategory = "board"
    directory_fmt = ["{category}", "{user}", "{board}"]
    pattern = [r"(?:https?://)?(?:[^./]+\.)?pinterest\.[^/]+/"
               r"(?!pin/)([^/?#&]+)/([^/?#&]+)"]
    test = [
        ("https://www.pinterest.com/g1952849/test-/", {
            "url": "85911dfca313f3f7f48c2aa0bc684f539d1d80a6",
            "keyword": "c54cf5aa830994f2ed4871efa7154a5fdaa1c2ce",
        }),
        ("https://www.pinterest.com/g1952848/test/", {
            "exception": exception.NotFoundError,
        }),
    ]

    def __init__(self, match):
        PinterestExtractor.__init__(self)
        self.user, self.board = match.groups()

    def items(self):
        board = self.api.board(self.user, self.board)
        data = self.data_from_board(board)
        num = data["count"]
        yield Message.Version, 1
        yield Message.Directory, data
        for pin in self.api.board_pins(self.user, self.board):
            url, pdata = self.data_from_pin(pin)
            data.update(pdata)
            data["num"] = num
            num -= 1
            yield Message.Url, url, data

    def data_from_board(self, board):
        """Get metadata from a board-object"""
        data = {
            "user": self.user,
            "board_id": util.safe_int(board["id"]),
            "board": board["name"],
            "count": board["counts"]["pins"],
        }
        return data


class PinterestPinitExtractor(PinterestExtractor):
    """Extractor for images from a pin.it URL"""
    subcategory = "pinit"
    pattern = [r"(?:https?://)?(pin\.it/[^/?#&]+)"]
    test = [
        ("https://pin.it/Hvt8hgT", {
            "url": "8daad8558382c68f0868bdbd17d05205184632fa",
        }),
        ("https://pin.it/Hvt8hgS", {
            "exception": exception.NotFoundError,
        }),
    ]

    def __init__(self, match):
        PinterestExtractor.__init__(self)
        self.url = "https://" + match.group(1)

    def items(self):
        response = self.session.head(self.url)
        location = response.headers.get("Location")
        if not location or location in ("https://api.pinterest.com/None",
                                        "https://www.pinterest.com"):
            raise exception.NotFoundError("pin")
        yield Message.Queue, location, {}


class PinterestAPI():
    """Minimal interface for the pinterest API"""

    def __init__(self, extractor, access_token=None):
        self.log = extractor.log
        self.session = extractor.session
        self.access_token = (
            access_token or
            extractor.config("access-token") or
            "AfyIXxi1MJ6et0NlIl_vBchHbex-FSWylPyr2GJE2uu3W8A97QAAAAA"
        )

    def pin(self, pin_id, fields="id,image,note"):
        """Query information about a pin"""
        endpoint = "pins/{}/".format(pin_id)
        params = {"fields": fields}
        return self._call(endpoint, params)["data"]

    def board(self, user, board, fields="id,name,counts"):
        """Query information about a board"""
        endpoint = "boards/{}/{}/".format(user, board)
        params = {"fields": fields}
        return self._call(endpoint, params)["data"]

    def board_pins(self, user, board, fields="id,image,note", limit=100):
        """Yield all pins of a specific board"""
        endpoint = "boards/{}/{}/pins/".format(user, board)
        params = {"fields": fields, "limit": limit}
        return self._pagination(endpoint, params)

    def _call(self, endpoint, params):
        params["access_token"] = self.access_token
        url = "https://api.pinterest.com/v1/" + endpoint

        response = self.session.get(url, params=params)
        status = response.status_code
        data = response.json()

        if 200 <= status < 400 and data.get("data"):
            return data

        msg = data.get("message", "")
        if status == 404:
            msg = msg.partition(" ")[0].lower()
            raise exception.NotFoundError(msg)
        self.log.error("API request failed: %s", msg or "")
        raise exception.StopExtraction()

    def _pagination(self, endpoint, params):
        while True:
            response = self._call(endpoint, params)
            yield from response["data"]

            cursor = response["page"]["cursor"]
            if not cursor:
                return
            params["cursor"] = cursor



# -*- coding: utf-8 -*-

# Copyright 2018 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extract manga-chapters and entire manga from https://mangadex.org/"""

from .common import ChapterExtractor, MangaExtractor
from .. import text, util, exception
from urllib.parse import urljoin
import json
import re


class MangadexExtractor():
    """Base class for mangadex extractors"""
    category = "mangadex"
    root = "https://mangadex.org"


class MangadexChapterExtractor(MangadexExtractor, ChapterExtractor):
    """Extractor for manga-chapters from mangadex.org"""
    archive_fmt = "{chapter_id}_{page}"
    pattern = [r"(?:https?://)?(?:www\.)?mangadex\.(?:org|com)/chapter/(\d+)"]
    test = [
        ("https://mangadex.org/chapter/122094", {
            "keyword": "b4c83fe41f125eae745c2e00d29e087cc4eb78df",
            "content": "7ab3bef5caccb62b881f8e6e70359d3c7be8137f",
        }),
        # oneshot
        ("https://mangadex.org/chapter/138086", {
            "count": 64,
            "keyword": "9b1b7292f7dbcf10983fbdc34b8cdceeb47328ee",
        }),
        # NotFoundError
        ("https://mangadex.org/chapter/1", {
            "exception": exception.NotFoundError,
        })
    ]

    def __init__(self, match):
        self.chapter_id = match.group(1)
        url = self.root + "/chapter/" + self.chapter_id
        ChapterExtractor.__init__(self, url)

    def get_metadata(self, page):
        if "title='Warning'" in page and " does not exist." in page:
            raise exception.NotFoundError("chapter")

        info    , pos = text.extract(page, '="og:title" content="', '"')
        manga_id, pos = text.extract(page, '/images/manga/', '.', pos)
        _       , pos = text.extract(page, ' id="jump_group"', '', pos)
        _       , pos = text.extract(page, ' selected ', '', pos)
        language, ___ = text.extract(page, " title='", "'", pos-100)
        group   , pos = text.extract(page, '>', '<', pos)

        info = text.unescape(info)
        match = re.match(
            r"(?:(?:Vol\. (\d+) )?Ch\. (\d+)([^ ]*)|(.*)) "
            r"\(([^)]+)\)",
            info)

        return {
            "manga": match.group(5),
            "manga_id": util.safe_int(manga_id),
            "volume": util.safe_int(match.group(1)),
            "chapter": util.safe_int(match.group(2)),
            "chapter_minor": match.group(3) or "",
            "chapter_id": util.safe_int(self.chapter_id),
            "chapter_string": info.replace(" - MangaDex", ""),
            "group": text.unescape(group),
            "lang": util.language_to_code(language),
            "language": language,
        }

    def get_images(self, page):
        dataurl , pos = text.extract(page, "var dataurl = '", "'")
        pagelist, pos = text.extract(page, "var page_array = [", "]", pos)
        server  , pos = text.extract(page, "var server = '", "'", pos)

        base = urljoin(self.root, server + dataurl + "/")

        return [
            (base + page, None)
            for page in json.loads(
                "[" + pagelist.replace("'", '"').rstrip(",") + "]"
            )
        ]


class MangadexMangaExtractor(MangadexExtractor, MangaExtractor):
    """Extractor for manga from mangadex.org"""
    pattern = [r"(?:https?://)?(?:www\.)?(mangadex\.(?:org|com)/manga/\d+)"]
    test = [
        ("https://mangadex.org/manga/2946/souten-no-koumori", {
            "count": ">= 1",
            "keywords": {
                "manga": "Souten no Koumori",
                "manga_id": 2946,
                "title": "Oneshot",
                "volume": 0,
                "chapter": 0,
                "chapter_minor": "",
                "chapter_id": int,
                "group": str,
                "contributor": str,
                "date": str,
                "views": int,
                "lang": str,
                "language": str,
            },
        }),
    ]
    scheme = "https"
    per_page = 100

    def chapters(self, page):
        results = []
        extr = text.extract
        num = 1

        manga = text.unescape(extr(
            page, '"og:title" content="', '"')[0].rpartition(" (")[0])
        manga_id = util.safe_int(extr(
            page, '/images/manga/', '.')[0])

        while True:
            before = len(results)

            for info in text.extract_iter(page, '<tr id="chapter_', '</tr>'):
                chid    , pos = extr(info, 'data-chapter-id="', '"')
                chapter , pos = extr(info, 'data-chapter-num="', '"', pos)
                volume  , pos = extr(info, 'data-volume-num="', '"', pos)
                title   , pos = extr(info, 'data-chapter-name="', '"', pos)
                language, pos = extr(info, " title='", "'", pos)
                group   , pos = extr(info, "<td>", "</td>", pos)
                user    , pos = extr(info, "<td>", "</td>", pos)
                views   , pos = extr(info, ">", "<", pos)
                date    , pos = extr(info, ' datetime="', '"', pos)

                chapter, sep, minor = chapter.partition(".")

                results.append((self.root + "/chapter/" + chid, {
                    "manga": manga,
                    "manga_id": util.safe_int(manga_id),
                    "title": text.unescape(title),
                    "volume": util.safe_int(volume),
                    "chapter": util.safe_int(chapter),
                    "chapter_minor": sep + minor,
                    "chapter_id": util.safe_int(chid),
                    "group": text.unescape(text.remove_html(group)),
                    "contributor": text.remove_html(user),
                    "views": util.safe_int(views),
                    "date": date,
                    "lang": util.language_to_code(language),
                    "language": language,
                }))

            if len(results) - before != self.per_page:
                return results

            num += 1
            page = self.request("{}/_/{}/".format(self.url, num)).text



# -*- coding: utf-8 -*-

# Copyright 2016-2018 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extract images from https://www.tumblr.com/"""

from .common import Extractor, Message
from .. import text, util, exception
import re
import time


def _original_image(url):
    match = re.match(
        r"https?://\d+\.media\.tumblr\.com"
        r"((/[0-9a-f]+)?/tumblr_[^/?&#.]+_)\d+(\.[0-9a-z]+)",
        url)

    if not match:
        return (url,)
    root = "https://s3.amazonaws.com/data.tumblr.com"
    path, key, ext = match.groups()

    return (
        "".join((root, path, "raw" if key else "1280", ext)),
        "".join((root, path, "500", ext)),
        url,
    )


def _original_video(url):
    return re.sub(
        (r"https?://vt\.media\.tumblr\.com"
         r"/tumblr_([^_]+)_\d+\.([0-9a-z]+)"),
        r"https://vt.media.tumblr.com/tumblr_\1.\2", url
    )


POST_TYPES = frozenset((
    "text", "quote", "link", "answer", "video", "audio", "photo", "chat"))

BASE_PATTERN = (
    r"(?:tumblr:(?:https?://)?([^/]+)|"
    r"(?:https?://)?([^.]+\.tumblr\.com))")


class TumblrExtractor(Extractor):
    """Base class for tumblr extractors"""
    category = "tumblr"
    directory_fmt = ["{category}", "{name}"]
    filename_fmt = "{category}_{blog_name}_{id}o{offset}.{extension}"
    archive_fmt = "{id}_{offset}"

    def __init__(self, match):
        Extractor.__init__(self)
        self.blog = match.group(1) or match.group(2)
        self.api = TumblrAPI(self)

        self.types = self._setup_posttypes()
        self.inline = self.config("inline", False)
        self.reblogs = self.config("reblogs", True)
        self.external = self.config("external", False)

        if len(self.types) == 1:
            self.api.posts_type = next(iter(self.types))
        elif not self.types:
            self.log.warning("no valid post types selected")

    def items(self):
        blog = None
        yield Message.Version, 1

        for post in self.posts():
            if post["type"] not in self.types:
                continue
            if not blog:
                blog = self.api.info(self.blog)
                yield Message.Directory, blog.copy()

            reblog = "reblogged_from_id" in post
            if reblog and not self.reblogs:
                continue
            post["reblogged"] = reblog

            post["blog"] = blog
            post["offset"] = 0

            if "trail" in post:
                del post["trail"]

            if "photos" in post:  # type "photo" or "link"
                photos = post["photos"]
                del post["photos"]

                for photo in photos:
                    post["photo"] = photo
                    photo.update(photo["original_size"])
                    del photo["original_size"]
                    del photo["alt_sizes"]
                    yield self._prepare_image(photo["url"], post)

            if "audio_url" in post:  # type: "audio"
                yield self._prepare(post["audio_url"], post)

            if "video_url" in post:  # type: "video"
                yield self._prepare(_original_video(post["video_url"]), post)

            if self.inline:  # inline images
                for key in ("body", "description", "source"):
                    if key in post:
                        for url in re.findall('<img src="([^"]+)"', post[key]):
                            yield self._prepare_image(url, post)

            if self.external:  # external links
                post["extension"] = None
                for key in ("permalink_url", "url"):
                    if key in post:
                        yield Message.Queue, post[key], post

    def posts(self):
        """Return an iterable containing all relevant posts"""

    def _setup_posttypes(self):
        types = self.config("posts", ("photo",))

        if types == "all":
            return POST_TYPES

        elif not types:
            return frozenset()

        else:
            if isinstance(types, str):
                types = types.split(",")
            types = frozenset(types)

            invalid = types - POST_TYPES
            if invalid:
                types = types & POST_TYPES
                self.log.warning('invalid post types: "%s"',
                                 '", "'.join(sorted(invalid)))
            return types

    @staticmethod
    def _prepare(url, post):
        text.nameext_from_url(url, post)
        post["offset"] += 1
        post["hash"] = post["name"].partition("_")[2]
        return Message.Url, url, post

    @staticmethod
    def _prepare_image(url, post):
        text.nameext_from_url(url, post)
        post["offset"] += 1

        parts = post["name"].split("_")
        post["hash"] = parts[1] if parts[1] != "inline" else parts[2]

        return Message.Urllist, _original_image(url), post


class TumblrUserExtractor(TumblrExtractor):
    """Extractor for all images from a tumblr-user"""
    subcategory = "user"
    pattern = [BASE_PATTERN + r"(?:/page/\d+)?/?$"]
    test = [
        ("http://demo.tumblr.com/", {
            "pattern": (r"https://s3\.amazonaws\.com/data\.tumblr\.com"
                        r"/tumblr_[^/_]+_\d+\.jpg"),
            "count": 1,
        }),
        ("http://demo.tumblr.com/", {
            "pattern": (r"https?://(?:$|"
                        r"s3\.amazonaws\.com/data\.tumblr\.com/.+_1280\.jpg|"
                        r"w+\.tumblr\.com/audio_file/demo/\d+/tumblr_\w+)"),
            "count": 3,
            "options": (("posts", "all"), ("external", True),
                        ("inline", True), ("reblogs", True))
        }),
        ("tumblr:http://www.b-authentique.com/", None),
        ("tumblr:www.b-authentique.com", None),
    ]

    def posts(self):
        return self.api.posts(self.blog, {})


class TumblrPostExtractor(TumblrExtractor):
    """Extractor for images from a single post on tumblr"""
    subcategory = "post"
    pattern = [BASE_PATTERN + r"/(?:post|image)/(\d+)"]
    test = [
        ("http://demo.tumblr.com/post/459265350", {
            "pattern": (r"https://s3\.amazonaws\.com/data\.tumblr\.com"
                        r"/tumblr_[^/_]+_1280.jpg"),
            "count": 1,
        }),
        ("http://demo.tumblr.com/image/459265350", None),
    ]

    def __init__(self, match):
        TumblrExtractor.__init__(self, match)
        self.post_id = match.group(3)
        self.reblogs = True

    def posts(self):
        return self.api.posts(self.blog, {"id": self.post_id})

    @staticmethod
    def _setup_posttypes():
        return POST_TYPES


class TumblrTagExtractor(TumblrExtractor):
    """Extractor for images from a tumblr-user by tag"""
    subcategory = "tag"
    pattern = [BASE_PATTERN + r"/tagged/([^/?&#]+)"]
    test = [("http://demo.tumblr.com/tagged/Times%20Square", {
        "pattern": (r"https://s3\.amazonaws\.com/data\.tumblr\.com"
                    r"/tumblr_[^/_]+_1280.jpg"),
        "count": 1,
    })]

    def __init__(self, match):
        TumblrExtractor.__init__(self, match)
        self.tag = text.unquote(match.group(3))

    def posts(self):
        return self.api.posts(self.blog, {"tag": self.tag})


class TumblrLikesExtractor(TumblrExtractor):
    """Extractor for images from a tumblr-user's liked posts"""
    subcategory = "likes"
    directory_fmt = ["{category}", "{name}", "likes"]
    archive_fmt = "f_{blog[name]}_{id}_{offset}"
    pattern = [BASE_PATTERN + r"/likes"]
    test = [("http://mikf123.tumblr.com/likes", {
        "count": 1,
    })]

    def posts(self):
        return self.api.likes(self.blog)


class TumblrAPI():
    """Minimal interface for the Tumblr API v2"""
    API_KEY = "O3hU2tMi5e4Qs5t3vezEi6L0qRORJ5y9oUpSGsrWu8iA3UCc3B"
    API_SECRET = "sFdsK3PDdP2QpYMRAoq0oDnw0sFS24XigXmdfnaeNZpJpqAn03"
    BLOG_CACHE = {}

    def __init__(self, extractor):
        api_key = extractor.config("api-key", self.API_KEY)
        api_secret = extractor.config("api-secret", self.API_SECRET)
        token = extractor.config("access-token")
        token_secret = extractor.config("access-token-secret")

        if api_key and api_secret and token and token_secret:
            self.session = util.OAuthSession(
                extractor.session,
                api_key, api_secret,
                token, token_secret,
            )
            self.api_key = None
        else:
            self.session = extractor.session
            self.api_key = api_key

        self.posts_type = None
        self.log = extractor.log

    def info(self, blog):
        """Return general information about a blog"""
        if blog not in self.BLOG_CACHE:
            self.BLOG_CACHE[blog] = self._call(blog, "info", {})["blog"]
        return self.BLOG_CACHE[blog]

    def posts(self, blog, params):
        """Retrieve published posts"""
        params.update({"offset": 0, "limit": 50, "reblog_info": "true"})
        if self.posts_type:
            params["type"] = self.posts_type
        while True:
            data = self._call(blog, "posts", params)
            self.BLOG_CACHE[blog] = data["blog"]
            yield from data["posts"]
            params["offset"] += params["limit"]
            if params["offset"] >= data["total_posts"]:
                return

    def likes(self, blog):
        """Retrieve liked posts"""
        params = {"limit": 50}
        while True:
            posts = self._call(blog, "likes", params)["liked_posts"]
            if not posts:
                return
            yield from posts
            params["before"] = posts[-1]["liked_timestamp"]

    def _call(self, blog, endpoint, params):
        if self.api_key:
            params["api_key"] = self.api_key
        url = "https://api.tumblr.com/v2/blog/{}/{}".format(
            blog, endpoint)

        response = self.session.get(url, params=params)
        data = response.json()
        status = data["meta"]["status"]

        if 200 <= status < 400:
            return data["response"]
        elif status == 403:
            raise exception.AuthorizationError()
        elif status == 404:
            raise exception.NotFoundError("user or post")
        elif status == 429:

            # daily rate limit
            if response.headers.get("x-ratelimit-perday-remaining") == "0":
                self.log.error(
                    "Daily API rate limit exceeded: aborting; "
                    "%s seconds until rate limit reset",
                    response.headers.get("x-ratelimit-perday-reset"),
                )
                raise exception.StopExtraction()

            # hourly rate limit
            reset = response.headers.get("x-ratelimit-perhour-reset")
            if reset:
                self.log.info(
                    "Hourly API rate limit exceeded; "
                    "waiting %s seconds for rate limit reset",
                    reset,
                )
                time.sleep(int(reset) + 1)
                return self._call(blog, endpoint, params)

        self.log.error(data)
        raise exception.StopExtraction()



# -*- coding: utf-8 -*-

# Copyright 2018 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extract manga-chapters and entire manga from https://subapics.com/"""

from . import komikcast
from .. import text


class SubapicsBase(komikcast.KomikcastBase):
    """Base class for subapics extractors"""
    category = "subapics"
    root = "https://subapics.com"


class SubapicsChapterExtractor(SubapicsBase,
                               komikcast.KomikcastChapterExtractor):
    """Extractor for manga-chapters from subapics.com"""
    pattern = [r"(?:https?://)?(?:www\.)?subapics\.com"
               r"(/[^/?&#]+-chapter-[^/?&#]+/?)$"]
    test = [("http://subapics.com/apotheosis-chapter-02-2/", {
        "url": "978d3c053d34a77f6ea6e60cbba3deda1e369be8",
        "keyword": "ed64479aef5a68aafa39334515f34a4595858c3c",
    })]

    @staticmethod
    def get_images(page):
        readerarea = text.extract(
            page, '<div id="readerarea">', '<meta/>')[0]
        return [
            (url, None)
            for url in text.extract_iter(
                readerarea, ' src="', '"'
            )
        ]


class SubapicsMangaExtractor(SubapicsBase,
                             komikcast.KomikcastMangaExtractor):
    """Extractor for manga from subapics.com"""
    pattern = [r"(?:https?://)?(?:www\.)?(subapics\.com/manga/[^/?&#]+/?)$"]
    test = [(("https://subapics.com/manga/"
              "rune-factory-4-koushiki-comic-visual-book/"), {
        "url": "6b18ba9513a6c92a23df1b78b11a1ad0013c6e5e",
        "keyword": "2fffe06b93b7ac8c4bb61f44398326deaf59fcf9",
    })]

    @staticmethod
    def get_metadata(page):
        manga , pos = text.extract(page, "<title>", "</title>")
        author, pos = text.extract(page, "<b>Author</b>: ", "</li>", pos)
        genres, pos = text.extract(page, "<b>Genres</b>: ", "</li>", pos)

        return {
            "manga": text.unescape(manga.rpartition(" – ")[0]),
            "author": text.unescape(author),
            "genres": text.remove_html(genres).replace(" , ", ", "),
        }
